#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_sparkonda
----------------------------------

Tests for `sparkonda` module.
"""
from __future__ import print_function
import unittest
from nose.tools import assert_true, assert_equal
from nose.plugins.attrib import attr
import sys
import ConfigParser
import os.path
from os.path import expanduser
import logging

from pyspark import SparkContext
from pyspark import SparkConf


class SparkondaTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):

        class_name = cls.__name__
        conf = SparkConf()
        conf.set('spark.app.name', 'class_name')

        # Read the spark configuration and update the spark conf
        test_spark_config = ConfigParser.ConfigParser()
        test_spark_config.read('test_config.cfg')
        test_spark_config.sections()
        configs = dict(test_spark_config.items('spark_conf_test_generic'))
        for k, v in configs.items():
            conf.set(k, v)
        cls.spark_test_configs = configs
        # Create the spark context
        cls.sc = SparkContext(conf=conf)
        if 'PYSPARK_DRIVER_PYTHON' in configs.keys():
            cls.sc.pythonExec = configs['PYSPARK_DRIVER_PYTHON']
        else:
            cls.sc.pythonExec = 'python2.7'

        logger = cls.sc._jvm.org.apache.log4j
        logger.LogManager.getLogger("org").setLevel(logger.Level.ERROR)
        logger.LogManager.getLogger("akka").setLevel(logger.Level.ERROR)

        logging.basicConfig(format='%(asctime)s %(name)s %(levelname)s: %(message)s')
        cls.logger = logging.getLogger(__name__)
        cls.logger.setLevel(logging.DEBUG)

    @classmethod
    def tearDownClass(cls):
        cls.sc.stop()
        # To avoid Akka rebinding to the same port, since it doesn't unbind
        # immediately on shutdown
        cls.sc._jvm.System.clearProperty("spark.driver.port")

    def test_01_spark_context_started(self):
        self.logger.debug("Running test_01_spark_context_started")
        assert_true(self.sc is not None, 'sc should be initialized')
        assert_true(self.sc.range(10).count() == 10, 'sc.range(10).count() should return 10')

    def test_02_adding_sparkonda_utils_file_to_workers(self):
        from os.path import expanduser
        home_dir = expanduser("~")
        self.logger.debug("Running test_02_adding_sparkonda_utils_file_to_workers")
        sparkonda_utils_filename = self.add_sparkonda_utils_to_workers()
        self.logger.debug('sparkonda_utils_filename: %s', sparkonda_utils_filename)
        assert_true('sparkonda' not in sys.modules.keys(),
                    'sparkonda should be cleaned from the sys.modules cache')
        assert_true('sparkonda.module_helper' not in sys.modules.keys(),
                    'sparkonda should be cleaned from the sys.modules cache')

        skon = self.import_sparkonda_utils()
        self.logger.debug("self.spark_test_configs['spark.executor.instances']: %s"
                          % str(self.spark_test_configs['spark.executor.instances']))
        self.logger.debug("self.spark_test_configs['spark.executor.cores']: %s"
                          % str(self.spark_test_configs['spark.executor.cores']))

        skon.SC_NUM_EXECUTORS = int(self.spark_test_configs['spark.executor.instances'])
        skon.SC_NUM_CORES_PER_EXECUTOR = int(self.spark_test_configs['spark.executor.cores'])
        skon.CONDA_ENV_NAME = 'sparkonda-test-env'
        skon.CONDA_ENV_LOCATION = ''.join([home_dir, '/miniconda/envs/', skon.CONDA_ENV_NAME])

        file_list = skon.list_cwd_files(self.sc, debug=True)
        self.logger.debug('file_list: %s' % str(file_list))
        assert_true(sparkonda_utils_filename in str(file_list),
                    ' '.join([sparkonda_utils_filename, ' should be in workers directories']))

    def test_03_pack_conda_env_on_the_spark_driver(self):
        self.logger.debug("Running test_03_pack_conda_env_on_the_spark_driver")
        skon = self.init_sparkonda()
        skon.pack_conda_env(overwrite=True)

        pack_location = '/tmp/' + skon.CONDA_ENV_NAME + '.tar'
        assert_true(pack_location is not None, 'Pack file location should not be None')
        assert_true('/tmp/' in pack_location, 'Pack file location should be in the /tmp directory')

        pack_location = skon.pack_conda_env(overwrite=False)
        assert_equal(pack_location, None, 'Pack file location should be None, if it already exist')

    def test_04_distribute_conda_env_on_the_spark_workers(self):
        self.logger.debug("Running test_04_distribute_conda_env_on_the_spark_workers")
        skon = self.init_sparkonda()
        skon.pack_conda_env(overwrite=True)
        pack_file_location = '/tmp/' + skon.CONDA_ENV_NAME + '.tar'
        skon.distribute_conda_env(self.sc)
        workers_files = skon.list_cwd_files(self.sc)
        self.logger.debug('workers_files:%s' % str(workers_files))
        pack_file_name = os.path.split(pack_file_location)[-1]
        count_files = 0
        for fh in workers_files:
            for f in fh:
                if pack_file_name == f:
                    count_files += 1
        self.logger.debug(
            'count_files, int(skon.SC_NUM_EXECUTORS): %s' % str([count_files, int(skon.SC_NUM_EXECUTORS)]))
        assert_equal(count_files, int(skon.SC_NUM_EXECUTORS),
                     'Number of pack files should be equal to the executors count')

    def test_05_install_conda_env_on_the_spark_workers(self):
        self.logger.debug("Running test_05_install_conda_env_on_the_spark_workers")
        skon = self.init_sparkonda()
        skon.pack_conda_env(overwrite=True)
        pack_file_location = '/tmp/' + skon.CONDA_ENV_NAME + '.tar'
        skon.distribute_conda_env(self.sc)
        workers_files = skon.list_cwd_files(self.sc)
        self.logger.debug('Before unpack workers_files:%s' % str(workers_files))
        skon.install_conda_env(self.sc)
        workers_files = skon.list_cwd_files(self.sc)
        self.logger.debug('After unpack workers_files:%s' % str(workers_files))
        self.logger.debug('Looking for top level folder in workers_files:%s' % skon.CONDA_ENV_LOCATION)
        top_level_conda_env_location = skon.CONDA_ENV_LOCATION.split('/')[1]
        count_files = 0
        for fh in workers_files:
            for f in fh:
                if top_level_conda_env_location == f:
                    count_files += 1
        self.logger.debug(
            'count_files, int(skon.SC_NUM_EXECUTORS): %s' % str([count_files, int(skon.SC_NUM_EXECUTORS)]))
        assert_equal(count_files, int(skon.SC_NUM_EXECUTORS),
                     'Number of conda env folders should be equal to the executors count')

    def test_06_conda_installed_python_on_the_spark_workers(self):
        self.logger.debug("Running test_06_conda_installed_python_on_the_spark_workers")
        skon = self.init_sparkonda()
        skon.pack_conda_env(overwrite=True)
        skon.distribute_conda_env(self.sc)
        skon.install_conda_env(self.sc)
        workers_files = skon.list_cwd_files(self.sc, debug=True)
        self.logger.debug('workers_files:%s' % str(workers_files))
        self.logger.debug(skon.CONDA_ENV_LOCATION)
        skon.set_workers_python_interpreter(self.sc)

        def check_pandas(x):
            import pandas as pd
            return [pd.__version__]

        pandas_versions = skon.prun(self.sc, check_pandas, include_broadcast_vars=False)
        self.logger.debug('pandas_versions %s' % str(pandas_versions))

        def check_sklearn(x):
            import sklearn as sk
            return [sk.__version__]

        sklearn_versions = skon.prun(self.sc, check_sklearn, include_broadcast_vars=False)
        self.logger.debug('sklearn_versions %s' % str(pandas_versions))

        assert_equal(len(pandas_versions), int(skon.SC_NUM_EXECUTORS),
                     'Number of responses should be equal to the executors count')

        assert_equal(len(sklearn_versions), int(skon.SC_NUM_EXECUTORS),
                     'Number of responses should be equal to the executors count')


    def test_07_num_executors_match_num_hosts_returned_from_prun(self):
        self.logger.debug("Running test_07_num_executors_match_num_hosts_returned_from_prun")
        skon = self.init_sparkonda()
        workers_files = skon.list_cwd_files(self.sc, debug=True)
        num_hosts = len(set(workers_files[::2]))
        assert_equal(num_hosts, int(skon.SC_NUM_EXECUTORS),
                     'Number of hosts should be equal to the executors count')


    def add_sparkonda_utils_to_workers(self):
        # Helper to add sparkonda_utils module to the workers
        # and clean up the sys.modules cache afterward
        import sparkonda
        return sparkonda.module_helper.add_module_to_workers(self.sc)

    def import_sparkonda_utils(self):
        # Helper to import the sparkonda_utils module
        # Try-Catch trick used for IDEs, to provide autocomplete
        try:
            import sparkonda_utils as skon
        except ImportError:
            import sparkonda.sparkonda_utils as skon
        return skon

    def init_sparkonda(self):
        from os.path import expanduser
        home_dir = expanduser("~")
        self.add_sparkonda_utils_to_workers()
        skon = self.import_sparkonda_utils()
        skon.SC_NUM_EXECUTORS = int(self.spark_test_configs['spark.executor.instances'])
        skon.SC_NUM_CORE_PER_EXECUTOR = int(self.spark_test_configs['spark.executor.cores'])
        skon.CONDA_ENV_NAME = 'sparkonda-test-env'
        skon.CONDA_ENV_LOCATION = ''.join([home_dir, '/miniconda/envs/', skon.CONDA_ENV_NAME])
        return skon



def d(x):
    from sklearn import datasets
    from sklearn.cross_validation import cross_val_predict
    from sklearn import linear_model
    lr = linear_model.LinearRegression()
    boston = datasets.load_boston()
    y = boston.target

    # cross_val_predict returns an array of the same size as `y` where each entry
    # is a prediction obtained by cross validated:
    predicted = cross_val_predict(lr, boston.data, y, cv=10)
    return predicted



# -*- coding: utf-8 -*-
import os
import tarfile
import logging
import subprocess

from pyspark import SparkFiles

"""
Name of the Conda environment to be used for the installation
"""
CONDA_ENV_NAME = None

"""
Location of the Conda environment on the driver node to be used for the installation
"""
CONDA_ENV_LOCATION = None

"""
Number of Executors used for the Spark job
"""
SC_NUM_EXECUTORS = 2

"""
Number of Cores per Executor used for the Spark job
"""
SC_NUM_CORES_PER_EXECUTOR = 2

"""Unpacking Error Level"""
UNPACKING_ERROR_LEVEL = 0

logging.basicConfig(format='%(asctime)s %(name)s %(levelname)s: %(message)s')
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


def __ls(broadcast_vars, iterator):
    """
    Get the list of files in the worker-local directory
    """
    return [__get_hostname(), os.listdir(SparkFiles.getRootDirectory())]


def __parse_worker_local_path(path):
    """
    Get the top level folder of the conda env after unzipping
    """
    return path[1:].split('/')[0]


def __unpack_conda_env(broadcast_vars, iterator):
    """
    Command to install the conda env in the local worker
    """
    # transform '/path/to/conda' to 'path'
    worker_conda_folder = __parse_worker_local_path(broadcast_vars['CONDA_ENV_LOCATION'])
    # delete existing unpacked env root 'path' folder
    cmd = 'rm -rf %s' % worker_conda_folder
    subprocess.check_output(cmd.split(' '))

    # unpack the env
    with tarfile.open(broadcast_vars['CONDA_ENV_NAME'] + '.tar') as tar_handle:
        tar_handle.errorlevel = UNPACKING_ERROR_LEVEL
        tar_handle.extractall()

    # remove the env tar file
    cmd = 'rm -rf %s.tar' % broadcast_vars['CONDA_ENV_NAME']
    subprocess.check_output(cmd.split(' '))

    return [__get_hostname(), 'done']


def __rm_conda_env(broadcast_vars, iterator):
    """
    Command to remove the conda env from the local worker
    """
    import subprocess
    worker_conda_folder = __parse_worker_local_path(broadcast_vars['CONDA_ENV_LOCATION'])

    rm_conda_cmd = 'rm -rf %s' % worker_conda_folder
    rm_conda_zip = 'rm -rf %s.tar' % broadcast_vars['CONDA_ENV_NAME']
    return [subprocess.check_output(rm_conda_cmd.split(' ')),
            subprocess.check_output(rm_conda_zip.split(' '))]


def __get_hostname():
    """
    Get the hostname of the worker
    """
    import socket
    return socket.gethostname()


def __get_broadcast_vars():
    """
    Build a dict of the broadcast vars for convenience
    """
    return dict(CONDA_ENV_NAME=CONDA_ENV_NAME,
                CONDA_ENV_LOCATION=CONDA_ENV_LOCATION,
                SC_NUM_EXECUTORS=SC_NUM_EXECUTORS)


def __get_local_module_file_location():
    """Discover the location of the sparkonda module
    """
    module_file = os.path.abspath(__file__).replace('pyc', 'py')
    return module_file


def __get_sc_executors_instances(sc):
    """Discover the location of the sparkonda module
    """
    sc.getConf()


def __tar_env():
    """Untar the conda env tar file
    """
    with tarfile.open('/tmp/' + CONDA_ENV_NAME + '.tar', 'w') as tar_handle:
        for root, dirs, files in os.walk(CONDA_ENV_LOCATION):
            for cur_file in files:
                tar_handle.add(os.path.join(root, cur_file))


def prun(sc, cmd, include_broadcast_vars=True, debug=False):
    """
    Run a function on all spark executors.
    The number of executors needs to be defined with SC_NUM_EXECUTORS
    """
    num_workers = SC_NUM_EXECUTORS
    num_cores_per_worker = SC_NUM_CORES_PER_EXECUTOR
    if debug:
        logger.debug('Number of executors/partitions:' + str(SC_NUM_EXECUTORS))

    if include_broadcast_vars:
        from functools import partial
        broadcast_vars = sc.broadcast(__get_broadcast_vars())
        return (sc.parallelize(range(num_workers*num_cores_per_worker))
                .repartition(num_workers)
                .mapPartitions(partial(cmd, broadcast_vars.value))
                .collect())
    else:
        return (sc.parallelize(range(num_workers*num_cores_per_worker))
                .repartition(num_workers)
                .mapPartitions(cmd)
                .collect())


def pack_conda_env(overwrite=True):
    """
    Pack the local conda env located at CONDA_ENV_LOCATION
    """
    if overwrite:
        print('Overwriting tar file if exists at:' + '/tmp/' + CONDA_ENV_NAME + '.tar')
        __tar_env()
    else:
        if not os.path.exists('/tmp/%s.tar' % CONDA_ENV_NAME):
            print('Tar file does not exist, creating new tar file at:' + '/tmp/' + CONDA_ENV_NAME + '.tar')
            __tar_env()


def distribute_conda_env(sc):
    """
    Distributes the conda env in tar format given a sparkcontext
    """
    return sc.addFile('/tmp/%s.tar' % CONDA_ENV_NAME)


def list_cwd_files(sc, debug=False):
    """
    List the files in the temporary directory of each executor given a sparkcontext
    """

    return prun(sc, __ls, debug=debug)


def install_conda_env(sc):
    """
    Unpack and install the conda package file in each executor given a sparkcontext
    """
    return prun(sc, __unpack_conda_env)


def remove_conda_env(sc):
    """
    Remove the conda env from each executor given a sparkcontext
    """
    return prun(sc, __rm_conda_env)


def set_workers_python_interpreter(sc):
    """
    Set the interpreter for each executor given a sparkcontext
    """
    # Remove the first path separator from the conda env location
    sc.pythonExec = CONDA_ENV_LOCATION[1:] + "/bin/python2.7"


def reset_workers_python_interpreter(sc):
    """
    Reset the interpreter to the default for each executor given a sparkcontext
    """
    sc.pythonExec = '/usr/bin/python2.7'



# # create a new conda env to test with if you don't have one:
# # conda create -n sparkonda-test python pandas scikit-learn
#
# #Add the sparkonda file to the cluster workers
"""
sc.addPyFile('dev/sparkonda/sparkonda/sparkonda.py')

import sparkonda as skon
try: import sparkonda.sparkonda as skon
except:print 'shell version using the distributed pyfile'
from os.path import expanduser
home_dir = expanduser("~")

#Edit to match your conda env name
skon.CONDA_ENV_NAME = 'sparkonda-test'
#Edit this path to match your conda env location
skon.CONDA_ENV_LOCATION = home_dir+'/miniconda/envs/'+skon.CONDA_ENV_NAME
#Edit to match your cluster size
skon.SC_NUM_EXECUTORS = 2

skon.zip_conda_env()
skon.distribute_conda_env(sc)
skon.list_cwd_files(sc)
skon.install_conda_env(sc)
skon.set_workers_python_interpreter(sc)

#This assumes that pandas and sklearn are installed in the conda env you specified
def d(x): import pandas as pd; return pd.__version__
sc.parallelize([1]).map(d).collect()

def d(x): import sklearn as sk; return sk.__version__
sc.parallelize([1]).map(d).collect()

skon.remove_conda_env(sc)

skon.list_cwd_files(sc)

sc.parallelize([1]).map(lambda x: x+1).collect()

#Check that the package is not accessible anymore(should get an error)
#ImportError: No module named sklearn
def d(x): import sklearn as sk; return sk.__version__

sc.parallelize([1]).map(d).collect()
"""
#
#
#



# -*- coding: utf-8 -*-

SPARKONDA_UTILS_FILENAME = 'sparkonda_utils.py'


def __get_local_module_file_location():
    """Discover the location of the sparkonda module
    """
    import os.path
    module_file = '/'.join([os.path.abspath(os.path.dirname(__file__)), SPARKONDA_UTILS_FILENAME])
    return module_file


def add_module_to_workers(sc, debug=False):
    """ Add the sparkonda module it to the workers and return the filename
    """
    import sys
    import os
    module_path = __get_local_module_file_location()
    if debug:
        print('Adding module to workers, module_path:', module_path)

    sc.addPyFile(module_path)

    del sys.modules['sparkonda']
    del sys.modules['sparkonda.module_helper']

    return os.path.split(module_path)[-1]
